import pytest
import asyncio
from .one_actor_app import app_for_data_provider

from fediverse_pasture.data_provider import DataProvider


@pytest.fixture
def client_one_actor():
    dp = DataProvider.generate(with_possible_actors=False)
    app = app_for_data_provider(dp)

    yield app.test_client()


async def test_inbox_display_no_content(client_one_actor):
    async with client_one_actor.request("/") as connection:
        data = await connection.receive()

        assert "Inbox content" in data.decode()

        await connection.disconnect()


async def test_inbox_display(client_one_actor):
    async with client_one_actor.request("/") as connection:
        data = await connection.receive()

        assert "Inbox content" in data.decode()
        await connection.disconnect()

    response = await client_one_actor.post("/actor/inbox", json={"test": "find me"})

    assert response.status_code == 202

    await asyncio.sleep(0.1)

    async with client_one_actor.request("/") as connection:
        data = await connection.receive()
        assert "Inbox content" in data.decode()
        data = await connection.receive()
        assert "find me" in data.decode()

        await connection.disconnect()



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import aiohttp
import click

from typing import Tuple
from contextlib import asynccontextmanager
from urllib.parse import urljoin

from bovine import BovineActor
from bovine.activitystreams import Actor

from .data_provider import (
    DataProvider,
    bovine_actor_for_actor_data,
)
from .server import assets_blueprint_for_directory
from .server.image_provider_blueprint import image_provider_blueprint
from .server.one_actor_app import app_for_data_provider
from .server.nodeinfo_blueprint import nodeinfo_blueprint

from .version import __version__


def bovine_actor_and_actor_object(domain: str) -> Tuple[BovineActor, Actor]:
    """Returns the BovineActor and the Actor object corresponding to the
    actor being served by one_actor

    :param domain: The domain the actor lives on
    :return: BovineActor and Actor object"""
    dp = DataProvider.load()

    actor_id = urljoin(domain, "/actor")

    return bovine_actor_for_actor_data(actor_id, dp.one_actor)


@asynccontextmanager
async def bovine_actor_and_session(
    domain: str = "http://pasture-one-actor",
):
    async with aiohttp.ClientSession() as session:
        bovine_actor, actor = bovine_actor_and_actor_object(domain)
        await bovine_actor.init(session=session)

        yield bovine_actor, actor, session


@click.command()
@click.option("--only_generate_config", default=False, is_flag=True)
@click.option(
    "--assets",
    default=None,
    help="Directory to serve assets from. Assets will be served under /assets/",
)
@click.option("--port", default=2909, help="port to run on")
@click.option("--reload", default=False, is_flag=True)
@click.option("--with_nodeinfo", default=False, is_flag=True)
def one_actor(only_generate_config, assets, port, reload, with_nodeinfo):
    print(f"fediverse pasture one_actor {__version__}")

    dp = DataProvider.generate_and_load(only_generate_config)

    app = app_for_data_provider(dp)

    app.register_blueprint(image_provider_blueprint, url_prefix="/images")

    if assets:
        app.register_blueprint(assets_blueprint_for_directory(assets))

    if with_nodeinfo:
        app.register_blueprint(nodeinfo_blueprint)

    app.run(port=port, host="0.0.0.0", use_reloader=reload)


if __name__ == "__main__":
    one_actor()



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from bovine import BovineActor
from bovine.activitystreams import Actor

from . import DataProvider, bovine_actor_for_actor_data


def test_actor_data():
    dp = DataProvider.generate()

    data = dp.one_actor.model_dump()

    assert data["actor_name"] == "actor"
    assert data["user_part"] == "actor"


def test_bovine_actor():
    dp = DataProvider.generate()

    actor, data = bovine_actor_for_actor_data("http://localhost/actor", dp.one_actor)

    assert isinstance(actor, BovineActor)
    assert isinstance(data, Actor)



# SPDX-FileCopyrightText: 2024 Helge
#
# SPDX-License-Identifier: MIT

import pytest

from .runner import ActivityRunner

from .helper import activity_runner


@pytest.mark.skip("Currently requires data.toml to exist")
async def test_activity_runner():
    async with activity_runner("mastodon") as runner:
        assert isinstance(runner, ActivityRunner)



from secrets import token_urlsafe
from collections.abc import Callable

from urllib.parse import urljoin


def create_make_id_for_actor_id(actor_id: str) -> Callable[..., str]:
    def make_id(activity: bool = False) -> str:
        prefix = "/activity/" if activity else "/object/"

        return urljoin(actor_id, f"{prefix}{token_urlsafe(8)}")

    return make_id



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import pytest
from .generate import generate_possible_actors


@pytest.mark.skip("SLOW")
def test_generate_possible_actors():
    result = generate_possible_actors()

    assert isinstance(result, list)
    assert len(result) > 0

    print(result)



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import asyncio
import logging
from aiohttp.client_exceptions import ClientResponseError
from dataclasses import dataclass, field
from datetime import datetime
from typing import Callable, List

from bovine import BovineActor
from bovine.activitystreams import (
    factories_for_actor_object,
)
from bovine.activitystreams.activity_factory import ActivityFactory
from bovine.activitystreams.object_factory import ObjectFactory
from bovine.activitystreams.utils import combine_as_list

from fediverse_pasture.types import ApplicationAdapterForLastActivity, MessageModifier

from .id_maker import create_make_id_for_actor_id

logger = logging.getLogger(__name__)


@dataclass
class ActivitySender:
    """The ActivitySender class serves as a way to coordinate the process
    of sending the same activity to multiple Fediverse applications.

    The basic process is

    * Create an Activity with a published timestamp stored in published
    * Send this activity to applications using `send`
    * Retrieve the result from said applications

    The usual way to create an ActivitySender is the `for_actor` method,
    i.e.

    ```python
    activity_sender = ActivitySender.for_actor(bovine_actor, actor_object)
    ```
    """

    activity_factory: ActivityFactory
    object_factory: ObjectFactory
    bovine_actor: BovineActor
    make_id: Callable

    published: datetime | None = field(default=None)
    note: dict | None = field(default=None)
    activity: dict | None = field(default=None)
    object_id: str | None = field(default=None)

    sleep_after_getting_inbox: bool = field(default=True)

    include_mention: bool = field(
        default=False,
        metadata={
            "description": "Set to True if the created Note should mention the target"
        },
    )

    def init_create_note(self, modifier: MessageModifier):
        """Sets activity to a Create for a Note. Here the Note is
        constructed from a skeleton by applying `modifier` to it.
        To successfully send the note to most applications, modifier
        should set the Note's content, i.e.

        ```python
        >>> from bovine.testing import actor
        >>> from unittest.mock import AsyncMock, Mock
        >>> actor_mock = Mock(id=actor["id"])
        >>> actor_mock.build.return_value = actor
        >>> sender = ActivitySender.for_actor(AsyncMock(), actor_mock)
        >>> sender.init_create_note(lambda x: {**x, "content": "text"})
        >>> sender.note
        {'type': 'Note',
            'attributedTo': 'http://actor.example',
            'to': ['as:Public'],
            'id': 'http://actor.example/object/...
            'published': '...
            'content': 'text'}

        ```

        This method can be used to create objects of other types
        by overriding "type".
        """
        self.object_id = self.make_id()

        note = self.object_factory.note(
            id=self.object_id,
            to={"as:Public"},
        ).build()
        del note["@context"]
        self.note = modifier(note)
        self.published = datetime.fromisoformat(note["published"].removesuffix("Z"))

    async def send(self, remote: str):
        """Sends the activity to the remote user

        :param remote: Actor URI of the remote user"""

        if self.note is None:
            return

        note = {
            **self.note,
        }
        note["to"] = note["to"] + [remote]

        if self.include_mention:
            mention = {"type": "Mention", "href": remote}
            note["tag"] = combine_as_list(note.get("tag", []), [mention])

        create = self.activity_factory.create(
            note, id=self.make_id(activity=True)
        ).build()

        create["@context"] = [
            create["@context"],
            {"Hashtag": "as:Hashtag", "sensitive": "as:sensitive"},
        ]
        self.activity = create

        try:
            result = await self.bovine_actor.get(remote)
            if not isinstance(result, dict) or "inbox" not in result:
                logger.warning(
                    "Remote %s does not have an inbox, cannot post activity", remote
                )
                return None

            remote_inbox = result["inbox"]

            if self.sleep_after_getting_inbox:
                #
                # See https://codeberg.org/helge/funfedidev/issues/138#issuecomment-1640700
                #
                await asyncio.sleep(0.1)

            return await self.bovine_actor.post(remote_inbox, create)
        except ClientResponseError as e:
            logger.warning("Posting to inbox of %s failed with %s", remote, e.status)
            logger.exception(e)
            return None

    @staticmethod
    def for_actor(bovine_actor, actor_object):
        """Initializes the Activity Sender object for a given BovineActor
        and the corresponding actor object"""
        activity_factory, object_factory = factories_for_actor_object(
            actor_object.build()
        )

        return ActivitySender(
            activity_factory=activity_factory,
            object_factory=object_factory,
            bovine_actor=bovine_actor,
            make_id=create_make_id_for_actor_id(actor_object.id),
        )


@dataclass
class ActivityRunner:
    """Coordinates sending an activity to many applications through an ActivitySender
    instances"""

    activity_sender: ActivitySender = field(
        metadata={"description": " an activity sender"}
    )
    applications: List[ApplicationAdapterForLastActivity] = field(
        metadata={"description": "list of applications to run against"}
    )

    wait_time: float = field(
        default=0.5,
        metadata={
            "description": """Time in seconds between trying to fetch the activity from remote servers"""
        },
    )

    tries: int = field(
        default=20,
        metadata={
            "description": """Number of tries to fetch activity from remote servers"""
        },
    )

    skip_fetch: bool = field(
        default=False, metadata={"description": """Skips fetching the result"""}
    )

    async def fetch_activity(self, application, object_id):
        for _ in range(self.tries):
            result = await application.fetch_activity(object_id)
            if result:
                return result
            await asyncio.sleep(self.wait_time)

    async def run_for_modifier(self, modifier: Callable[[dict], dict]):
        """modifier has the same format as for ActivitySender

        :param modifier: modifies the base object being send"""
        self.activity_sender.init_create_note(modifier)

        async with asyncio.TaskGroup() as tg:
            for application in self.applications:
                tg.create_task(self.activity_sender.send(application.actor_uri))

        if self.skip_fetch:
            return

        await asyncio.sleep(self.wait_time)

        result = {"activity": self.activity_sender.activity}

        for application in self.applications:
            result[application.application_name] = await self.fetch_activity(
                application, self.activity_sender.object_id
            )

        return result



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import pytest

from quart import Quart

from fediverse_pasture.data_provider import DataProvider

from . import BlueprintForActors


@pytest.fixture
def client_one_actor():
    dp = DataProvider.generate(with_possible_actors=False)
    bfa = BlueprintForActors(actors=[dp.one_actor])

    app = Quart(__name__)
    app.register_blueprint(bfa.blueprint)

    yield app.test_client()


async def test_actor_webfinger(client_one_actor):
    response = await client_one_actor.get(
        "/.well-known/webfinger?resource=acct:actor@localhost"
    )

    assert response.status_code == 200


async def test_actor_get(client_one_actor):
    response = await client_one_actor.get("/actor")

    assert response.status_code == 200

    data = await response.get_json()

    assert isinstance(data, dict)


async def test_actor_not_found(client_one_actor):
    response = await client_one_actor.get("/other")

    assert response.status_code == 404


async def test_actor_post(client_one_actor):
    response = await client_one_actor.post("/actor/inbox", data={})

    assert response.status_code == 202



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import asyncio
import aiohttp
import logging
import secrets
import json
from typing import List
from urllib.parse import urlparse
from dataclasses import dataclass, field

from bovine.clients import lookup_uri_with_webfinger

from fediverse_pasture.data_provider import bovine_actor_for_actor_data
from fediverse_pasture.data_provider.models import ActorData
from fediverse_pasture.types import Message

logger = logging.getLogger(__name__)

remote_urls = {
    "documentation": "https://funfedi.dev/python_package/verify_actor/",
    "support_table": "https://funfedi.dev/support_tables/generated/verify_actor/",
    "repository": "https://codeberg.org/helge/funfedidev/",
}


def can_be_resolved(uri: str) -> bool:
    if uri.startswith("http://"):
        return True

    if uri.startswith("https://"):
        return True

    return False


@dataclass
class ActorVerifier:
    """Class implementing the logic to verify a remote actor"""

    actor_list: List[ActorData] = field(
        metadata={"description": "List of actors to run verification with"}
    )
    remote_uri: str = field(
        metadata={
            "description": "URI of the remote actor. If acct-uri, it is resolved using webfinger"
        }
    )
    domain: str = field(
        metadata={"description": "The domain the verification is done from"}
    )
    message: Message = field(
        metadata={"description": "Message object containing an event log"}
    )

    remote_actor_uri: str | None = None
    scheme: str = "http"

    timeout: int = field(
        default=20, metadata={"description": "The timeout for requests (in seconds)"}
    )

    async def verify(self, only: dict = {}):
        """Public interface, this method loops over the actors in `actor_list`
        and verifies using them if the remote_uri is accessible.

        Returns a dictionary with actor.actor_name as key and the
        result as value.

        This function creates it's own aiohttp.ClientSession"""
        async with aiohttp.ClientSession() as session:
            await self.determine_remote_actor_uri(session)
            result = {
                actor.actor_name: await self.verify_for_actor(session, actor)
                for actor in self.actor_list
                if actor.actor_name != "actor" and only.get(actor.actor_name, True)
            }
            result["webfinger"] = await self.check_webfinger_result(session)

            return result

    @property
    def main_actor(self) -> ActorData | None:
        for actor in self.actor_list:
            if actor.actor_name == "actor":
                return actor
        return None

    async def init_bovine_actor(self, actor, session):
        actor_id = f"{self.scheme}://{self.domain}/{actor.actor_name}"
        bovine_actor, _ = bovine_actor_for_actor_data(actor_id, actor)
        await bovine_actor.init(session=session)

        return bovine_actor

    async def check_webfinger_result(self, session):
        bovine_actor = await self.init_bovine_actor(self.main_actor, session)

        try:
            async with asyncio.timeout(self.timeout):
                actor = await bovine_actor.get(self.remote_actor_uri)
        except Exception as e:
            logger.exception(e)
            return False

        if actor is None:
            self.message.error("Failed to fetch actor")
            return False

        remote_actor_uri = actor.get("id", self.remote_actor_uri)

        preferred_username = actor.get("preferredUsername")

        self.message.add(f"Got preferredUsername {preferred_username}")
        if preferred_username is None:
            return False

        domain = urlparse(self.remote_actor_uri).netloc
        acct_uri = f"acct:{preferred_username}@{domain}"

        self.message.add(f"computed acct uri {acct_uri}")

        try:
            object_id, _ = await lookup_uri_with_webfinger(
                session, acct_uri, f"{self.scheme}://{domain}"
            )
        except Exception as e:
            logger.exception(e)
            return False
        self.message.add(f"Retrieved id {object_id} using webfinger")

        if object_id == remote_actor_uri:
            self.message.add("webfinger result matches expectations")

            return True

        return False

    async def verify_for_actor(self, session, actor):
        bovine_actor = await self.init_bovine_actor(actor, session)
        self.message.add(f"Running verification for {actor.actor_name}")

        if actor.requires_signed_post_for_inbox:
            return await self.fetch_remote_and_post_using_actor(bovine_actor)
        else:
            return await self.fetch_remote_and_post_using_session(session)

    async def fetch_remote_and_post_using_actor(self, bovine_actor):
        result = {"get_actor": False, "post_inbox": False}

        try:
            async with asyncio.timeout(self.timeout):
                actor = await bovine_actor.get(self.remote_actor_uri)
                if actor is None:
                    self.message.add(
                        f"Failed to retrieve actor {self.remote_actor_uri}"
                    )
                    return result

                inbox = actor.get("inbox")
                self.message.add(f"Got inbox {inbox}")

                if inbox:
                    result["get_actor"] = True
                    try:
                        try:
                            response = await bovine_actor.post(
                                inbox,
                                data={
                                    "@context": "https://www.w3.org/ns/activitystreams",
                                    "type": "Like",
                                    "actor": bovine_actor.actor_id,
                                    "id": bovine_actor.actor_id
                                    + secrets.token_urlsafe(8),
                                    "object": bovine_actor.actor_id,
                                },
                            )
                            self.message.add("Successfully posted to inbox with result")
                            self.message.add(response.status)
                            self.message.add((await response.text()))

                        except aiohttp.ClientResponseError as e:
                            if e.status == 400:
                                self.message.add(
                                    "Successfully posted to inbox but remote server\
    indicated a bad request"
                                )
                            else:
                                raise e

                        result["post_inbox"] = True
                    except Exception as e:
                        self.message.add("Failed to post to inbox")
                        self.message.add(repr(e))

        except Exception as e:
            self.message.add("Something went wrong")
            self.message.add(repr(e))

        return result

    async def resolve_inbox_using_session(self, session) -> str:
        async with asyncio.timeout(self.timeout):
            async with session.get(
                self.remote_actor_uri,
                headers={"accept": "application/activity+json"},
            ) as response:
                actor = json.loads(await response.text())
            inbox = actor.get("inbox")
            self.message.add(f"Got inbox {inbox}")
            return inbox

    async def fetch_remote_and_post_using_session(self, session):
        result = {"get_actor": False, "post_inbox": False}

        try:
            inbox = await self.resolve_inbox_using_session(session)
        except Exception as e:
            self.message.add("Something went wrong when fetching actor")
            self.message.add(repr(e))
            return result

        if inbox:
            result["get_actor"] = True
            try:
                async with asyncio.timeout(self.timeout):
                    async with session.post(
                        inbox,
                        data=json.dumps(
                            {
                                "@context": "https://www.w3.org/ns/activitystreams",
                                "type": "EchoRequest",
                            }
                        ),
                        headers={"content-type": "application/activity+json"},
                    ) as response:
                        self.message.add(f"Got {response.status} for unsigned POST")

                        if response.status < 400 and response.status > 100:
                            result["post_inbox"] = True
            except Exception as e:
                self.message.add("Something went wrong when posting to inbox")
                self.message.add(repr(e))
                return result

        return result

    async def determine_remote_actor_uri(self, session) -> None:
        """Used to resolve an acct-URI into the corresponding http/https-URI
        using webfinger"""
        if can_be_resolved(self.remote_uri):
            self.remote_actor_uri = self.remote_uri
            self.message.add(f"Can fetch actor from {self.remote_uri}")
            return

        self.message.add(f"Need to resolve {self.remote_uri} to actor object id")

        if self.remote_uri.startswith("acct:"):
            acct_uri = self.remote_uri
        else:
            self.message.add("Not in account uri format")

            if self.remote_uri[0] == "@":
                acct_uri = "acct:" + self.remote_uri[1:]
            else:
                acct_uri = "acct:" + self.remote_uri

        if "@" not in acct_uri:
            self.message.add(f"Computed invalid account URI {acct_uri}")
            return
        domain = acct_uri.split("@")[1]

        self.message.add(f"Resolving {acct_uri} using webfinger")

        try:
            object_id, _ = await lookup_uri_with_webfinger(
                session, acct_uri, f"{self.scheme}://{domain}"
            )
        except Exception as e:
            logger.exception(e)
            return

        self.message.add(f"Resolved to {object_id}")
        self.remote_actor_uri = object_id



import json

from bovine.clients import lookup_uri_with_webfinger
from bovine.utils import parse_fediverse_handle

from fediverse_pasture.runner import ActivitySender
from fediverse_pasture.one_actor import bovine_actor_and_session
from fediverse_pasture.types import MessageModifier


async def handle_send_to(
    modifier: MessageModifier, domain: str, uri: str, mention: bool, verbose: bool
):
    async with bovine_actor_and_session(domain) as (bovine_actor, actor, session):
        if uri.startswith("acct:"):
            _, acct_domain = parse_fediverse_handle(uri)

            candidate_uri, _ = await lookup_uri_with_webfinger(
                session, uri, domain=f"http://{acct_domain}"
            )

            if not candidate_uri:
                raise ValueError(f"Could not resolve {uri} to an actor URI")

            uri = candidate_uri

        sender = ActivitySender.for_actor(bovine_actor, actor)
        sender.include_mention = mention
        sender.init_create_note(modifier)

        result = await sender.send(uri)

        if verbose:
            print(json.dumps(sender.activity, indent=2))

        return result



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import pytest

from quart import Quart

from fediverse_pasture.data_provider import DataProvider

from . import BlueprintForActors


@pytest.fixture
def client_one_actor():
    dp = DataProvider.generate(with_possible_actors=False)
    actor = dp.one_actor
    actor.requires_signed_get_for_actor = True
    actor.requires_signed_post_for_inbox = True
    bfa = BlueprintForActors(actors=[actor])

    app = Quart(__name__)
    app.register_blueprint(bfa.blueprint)

    yield app.test_client()


async def test_actor_get_unsigned(client_one_actor):
    response = await client_one_actor.get("/actor")

    assert response.status_code == 401


async def test_actor_post_unsigned(client_one_actor):
    response = await client_one_actor.post("/actor/inbox", data={})

    assert response.status_code == 401



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import logging
import click
from quart import Quart


from .server.http_signature import http_signature_blueprint

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@click.command()
@click.option("--port", default=2909, help="port to run on")
@click.option("--reload", default=False, is_flag=True)
def http_signature(port, reload):
    app = Quart(__name__)
    app.register_blueprint(http_signature_blueprint)
    app.run(port=port, host="0.0.0.0", use_reloader=reload)


if __name__ == "__main__":
    http_signature()



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import aiohttp
from dataclasses import dataclass
from datetime import datetime
import bovine
import logging


from fediverse_pasture.types import ApplicationAdapterForLastActivity

logger = logging.getLogger(__name__)


@dataclass
class FirefishApplication:
    """Used to query a firefish application"""

    domain: str
    username: str
    session: aiohttp.ClientSession | None = None
    token: str = "VvGrKxhpIzJ1PomJFTBObYWOELgGniVi"

    firefish_id: str | None = None

    async def determine_actor_uri(self):
        actor_uri, _ = await bovine.clients.lookup_uri_with_webfinger(
            self.session, f"acct:{self.username}@{self.domain}", f"http://{self.domain}"
        )
        return actor_uri

    async def determine_actor_firefish_id(self):
        """Determines the firefish id for the user on the
        one_actor server, i.e. the one with hostname
        `pasture_one_actor"""
        response = await self.session.post(
            f"http://{self.domain}/api/users/search", data={"query": "actor"}
        )
        users = await response.json()

        for user in users:
            if user.get("host") == "pasture_one_actor":
                return user.get("id")

    async def user_post_with_published(self, published: datetime):
        if not self.firefish_id:
            self.firefish_id = await self.determine_actor_firefish_id()

        response = await self.session.post(
            f"http://{self.domain}/api/users/notes",
            data={"userId": self.firefish_id},
            headers={"Authorization": f"Bearer {self.token}"},
        )
        notes = await response.json()

        for data in notes:
            created_at = data.get("createdAt")
            created_at = datetime.fromisoformat(created_at.removesuffix("Z"))
            if created_at == published:
                return data

        return None

    async def top_public(self):
        response = await self.session.post(
            f"http://{self.domain}/api/notes/global-timeline"
        )
        public_timeline = await response.json()
        return public_timeline[0]

    async def top_public_with_published(self, published: datetime) -> dict | None:
        data = await self.top_public()
        created_at = data.get("createdAt")
        created_at = datetime.fromisoformat(created_at.removesuffix("Z"))
        if created_at == published:
            return data
        return None

    async def last_activity(
        self, session: aiohttp.ClientSession
    ) -> ApplicationAdapterForLastActivity:
        self.session = session

        actor_uri = await self.determine_actor_uri()

        return ApplicationAdapterForLastActivity(
            actor_uri=actor_uri,
            fetch_activity=self.user_post_with_published,
            application_name="firefish",
        )



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from unittest.mock import patch

from . import activity_for_mastodon, activity_for_firefish

from fediverse_pasture.types import ApplicationAdapterForLastActivity


async def test_activity_for_mastodon():
    result = await activity_for_mastodon("domain", "user", "xxx", "session")

    assert isinstance(result, ApplicationAdapterForLastActivity)


@patch("bovine.clients.lookup_uri_with_webfinger")
async def test_activity_for_firefish(mock_lookup):
    mock_lookup.return_value = "actor_uri", 0
    result = await activity_for_firefish("domain", "user", "session")

    assert isinstance(result, ApplicationAdapterForLastActivity)

    assert result.actor_uri == "actor_uri"



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from bovine.crypto import generate_rsa_public_private_key


from .models import ActorData, ActorKeyPair


def generate_key_pair(name: str) -> ActorKeyPair:
    public, private = generate_rsa_public_private_key()

    return ActorKeyPair(name=name, public=public, private=private)


def generate_one_actor():
    key_pair = generate_key_pair("main")

    return ActorData(actor_name="actor", user_part="actor", key_pairs=[key_pair])


def generate_possible_actors():
    result = []

    for (
        name,
        user_part,
        requires_signed_get_for_actor,
        requires_signed_post_for_inbox,
    ) in [
        ("alice", "alice", False, False),
        ("bob", "bob", False, True),
        ("claire", "claire", True, True),
        ("dean", None, False, False),
        ("emily", None, False, True),
        ("frank", None, True, True),
    ]:
        key_pair = generate_key_pair("main")

        summary = f"""user_part={user_part}
requires_signed_get_for_actor={requires_signed_get_for_actor},
requires_signed_post_for_inbox={requires_signed_post_for_inbox},
"""

        result.append(
            ActorData(
                actor_name=name,
                summary=summary,
                user_part=user_part,
                key_pairs=[key_pair],
                requires_signed_get_for_actor=requires_signed_get_for_actor,
                requires_signed_post_for_inbox=requires_signed_post_for_inbox,
            )
        )

    return result



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from .entry import Entry


def test_from_result_list_empty():
    entry = Entry.from_result_list([])

    assert isinstance(entry, Entry)
    assert not entry.present_for("app")
    assert entry.applications == set()


def test_from_result_list_one_element():
    result = [{"application_name": "app", "data": "boo"}]
    entry = Entry.from_result_list(result)

    assert isinstance(entry, Entry)
    assert entry.present_for("app")
    assert entry.applications == {"app"}

    lines = entry.as_tabs(["app", "other"])

    expected = """=== "app"

    ```json title="app"
    {
      "data": "boo"
    }
    ```


=== "other"

    no result

"""

    assert "".join(lines) == expected


def test_from_result_list_one_element_as_grid():
    result = [{"application_name": "app", "data": "boo"}]
    entry = Entry.from_result_list(result)

    lines = entry.as_grid(["app", "other"])

    expected = """<div class="grid" markdown>

```json title="app"
{
  "data": "boo"
}
```


no result

</div>
"""

    assert "".join(lines) == expected



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from quart import request, Blueprint


from bovine.crypto.helper import content_digest_sha256
from .utils import (
    validate_basic_signature_fields,
    retrieve_public_key,
    build_message_signature,
    build_http_signature,
    validate_signature,
)


http_signature_blueprint = Blueprint("http_signature", __name__)


@http_signature_blueprint.get("/")
async def get_handler():
    message, signature, error_dict = build_message_signature(request, "get")

    if error_dict:
        return error_dict, 401

    error_dict = validate_basic_signature_fields(message, request, signature)

    if error_dict:
        return error_dict, 401

    http_signature = build_http_signature(message, signature, request, "get")
    public_key, error_dict = await retrieve_public_key(
        message, signature.key_id, request
    )

    if error_dict:
        return error_dict, 401

    error_dict = validate_signature(
        message, request, public_key, signature, http_signature
    )

    if error_dict:
        return error_dict, 401

    return message.response, 200, {"content-type": "application/json"}


@http_signature_blueprint.post("/")
async def post_handler():
    message, signature, error_dict = build_message_signature(request, "post")

    if error_dict:
        return error_dict, 401

    body = await request.get_data()
    body_str = body.decode("utf-8")
    message.add(f"Got body: '{body_str}'")
    digest = content_digest_sha256(body)
    message.add(f"Computed digest {digest}")

    error_dict = validate_basic_signature_fields(
        message, request, signature, digest=digest
    )

    if error_dict:
        return error_dict, 401

    http_signature = build_http_signature(message, signature, request, "post")
    public_key, error_dict = await retrieve_public_key(
        message, signature.key_id, request
    )

    if error_dict:
        return error_dict, 401

    error_dict = validate_signature(
        message, request, public_key, signature, http_signature
    )

    if error_dict:
        return error_dict, 401

    return message.response, 202, {"content-type": "application/json"}



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

from pydantic import BaseModel, Field
from typing import List


class ActorKeyPair(BaseModel):
    """Represents a key pair for the actor"""

    name: str = Field(
        description="""Name of the key used in the key id in the form `key_id = f"{actor_id}#{name}"`"""
    )
    public: str = Field(description="""The PEM encoded public key""")
    private: str = Field(description="""The PEM encoded private key""")


class ActorData(BaseModel):
    """Represents an Actor"""

    actor_name: str = Field(
        description="""The name of the actor used in the actor_id"""
    )
    key_pairs: List[ActorKeyPair] = Field([], description="""List of keys""")
    user_part: str | None = Field(
        None,
        description="""User as part of the acct-uri for webfinger, None means webfinger lookup is not possible""",
    )

    summary: str = Field("", description="""Summary part of actor profile""")

    requires_signed_get_for_actor: bool = Field(
        False, description="""If true, validates the signature on `GET /actor`"""
    )
    requires_signed_post_for_inbox: bool = Field(
        False, description="""If true, validates the signature on `POST /inbox`"""
    )



# SPDX-FileCopyrightText: 2024 Helge
#
# SPDX-License-Identifier: MIT

from contextlib import asynccontextmanager

from fediverse_pasture.runner import (
    ActivitySender,
    ActivityRunner,
)
from fediverse_pasture.runner.application import app_name_to_coroutine
from fediverse_pasture.one_actor import bovine_actor_and_session


@asynccontextmanager
async def activity_runner(
    application_name: str,
):
    application_generator = app_name_to_coroutine.get(application_name)

    if application_generator is None:
        available_applications = ", ".join(app_name_to_coroutine.keys())
        raise ValueError(
            f"Unknown application '{application_name}' available applications are {available_applications}"
        )

    async with bovine_actor_and_session() as (bovine_actor, actor, session):
        sender = ActivitySender.for_actor(bovine_actor, actor)

        application = await application_generator(session)

        activity_runner = ActivityRunner(sender, [application])

        yield activity_runner



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

import logging
from datetime import datetime
from dataclasses import dataclass, field
from typing import List, Callable, Awaitable

logger = logging.getLogger(__name__)


@dataclass
class Message:
    """Class to track messages"""

    steps: List[str] = field(
        default_factory=list, metadata={"description": "a log of strings"}
    )

    def add(self, msg):
        """Adds a message to steps and logs it to info"""
        logger.info(msg)
        self.steps.append(msg)

    def error(self, msg):
        """Logs an error and returns a message object"""
        logger.error(msg)
        return {"x error": msg, **self.response}

    @property
    def response(self):
        """Returns a dictionary with a single key "steps" containing the steps"""
        return {"steps": self.steps}


@dataclass
class ApplicationAdapterForLastActivity:
    """Basic type that is used to describe how to interact with
    an external application. actor_uri represents the actor
    a message will be sent to. fetch_activity is used to
    fetch this activity.
    """

    actor_uri: str = field(metadata={"description": "The actor uri"})
    application_name: str = field(
        metadata={"description": "The name the application will be displayed as"}
    )
    fetch_activity: Callable[[datetime], Awaitable[dict | None]] = field(
        metadata={
            "description": "coroutine that retrieves the activity with a specified published date."
        }
    )


@dataclass
class ApplicationAdapterForActor:
    """Basic type that is used to describe how to interact with
    an external application. actor_uri represents the actor
    a message will be sent to.
    """

    actor_uri: str = field(metadata={"description": "The actor uri"})
    application_name: str = field(
        metadata={"description": "The name the application will be displayed as"}
    )


MessageModifier = Callable[[dict], dict]
"""Used to add the variable content to an activity pub message. The intended usage can be seen in the
[ActivitySender.init_create_note][fediverse_pasture.runner.ActivitySender.init_create_note] method.
"""
